/*
 MCL
 Copyright (c) 2012-18, Enzo De Sena
 All rights reserved.
 
 Authors: Enzo De Sena, enzodesena@me.com
 */

// This file contains definitions of matrix operations and classes

#ifndef MCL_MATRIXOP_H
#define MCL_MATRIXOP_H

#include <fstream>
#include <iostream>
#include <iomanip>
#include "mcltypes.h"
#include "comparisonop.h"
#include "basicop.h"
#include "elementaryop.h"

#if MCL_LOAD_EIGEN
  #include <Eigen/Dense>
#endif

#ifdef MCL_EXPORTS
  #define MCL_API __declspec(dllexport)
#else
  #define MCL_API
#endif

namespace mcl {

// Forward declaration
MCL_API std::vector<std::string> Split(const std::string& string, char delim);
  
/** Matrix class */
template<class T>
class MCL_API Matrix {
public:
  /** Default constructor with empty (0x0) matrix */
  Matrix() : num_rows_(0), num_columns_(0) {}
  
  /** 
   Constructs a matrix with default entries
   */
  Matrix(UInt num_rows, UInt num_columns) : num_rows_(num_rows), 
          num_columns_(num_columns) {
    for (UInt i=0; i<num_rows; ++i) {
      data_.push_back(std::vector<T>(num_columns));
    }
  }
  
  /**
   Constructs a matrix from a vector of vectors (outer vector represents rows)
   */
  Matrix(const std::vector<std::vector<T> > vectors) {
    num_rows_ = vectors.size();
    if (num_rows_ > 0) {
      num_columns_ = vectors[0].size();
      for (UInt i=1; i<num_rows_; ++i) {
        // Check that all rows have the same number of columns
        if (vectors[i].size() != num_columns_) { throw_line(""); }
      }
      data_ = vectors;
    }
    else {
      num_columns_ = 0;
    }
  }
  
  /** Sets element in given row and column */
  void set_element(const UInt& index_row, const UInt& index_column,
                   const T& element) {
    if (index_row>=num_rows_) { throw_line(""); }
    if (index_column>=num_columns_) { throw_line(""); }
    data_[index_row][index_column] = element;
  }
  
  /** Sets an entire column */
  void set_column(UInt index_column, std::vector<T> column) {
    if (column.size() != num_rows_) { throw_line(""); }
    if (index_column >= num_columns_) { throw_line(""); }
    for (UInt i=0; i<num_rows_; ++i) {
      set_element(i, index_column, column[i]);
    }
  }
  
  /** Sets an entire row */
  void set_row(UInt index_row, std::vector<T> row) {
    if (row.size() != num_columns_) { throw_line(""); }
    if (index_row >= num_rows_) { throw_line(""); }
    data_[index_row] = row;
  }
  
  /** Accesses an element in given row and column */
  T element(const UInt& index_row, const UInt& index_column) const {
    if (index_row >= num_rows_) { throw_line(""); }
    if (index_column >= num_columns_) { throw_line(""); }
    return data_[index_row][index_column];
  }
  
  /** Accesses an entire row */
  std::vector<T> row(UInt index_row) const {
    if (index_row >= num_rows_) { throw_line(""); }
    return data_[index_row];
  }
  
  /** Accesses an entire column */
  std::vector<T> column(UInt index_column) const {
    if (index_column >= num_columns_) { throw_line(""); }
    std::vector<T> output(num_rows_);
    for (UInt i=0; i<num_rows_; ++i) { output[i] = data_[i][index_column]; }
    return output;
  }
  
  /** Returns the serialised matrix. Equivalent to Matlab's matrix(:) */
  std::vector<T> Serial() const {
    std::vector<T> serial(num_columns()*num_rows());
    
    UInt k=0;
    for (UInt j=0; j<num_columns(); ++j) {
      for (UInt i=0; i<num_rows(); ++i) {
        serial[k++] = element(i, j);
      }
    }
    return serial;
  }
  
  /** Returns the number of rows */
  UInt num_rows() const { return num_rows_; }
  
  /** Returns the number of columns */
  UInt num_columns() const { return num_columns_; }
  
  /** Writes the matrix to a file. The optional parameter `precision` sets
   the number of decimal positions in the output file*/
  void Save(std::string file_name, mcl::UInt precision = 5) {
    std::ofstream output_file;
    output_file.open(file_name.c_str());
    output_file<<std::fixed;
    output_file<<std::setprecision((int)precision);
    for (UInt i=0; i<num_rows_; ++i) {
      for (UInt j=0; j<num_columns_; ++j) {
        output_file<<data_.at(i).at(j)<<" ";
      }
      output_file<<std::endl;
    }
    output_file.close();
  }
  
  /** Returns the raw data */
  std::vector<std::vector<T> > data() { return data_; }
  
  /**
   Reads a matrix. Elements have to be separated by tabs and there
   must be no ending empty line (e.g. 5 lines == 5 rows).
   */
  static Matrix Load(std::string file_name) {
    std::string line;
    std::ifstream in_file (file_name.c_str());
    if (! in_file.is_open()) { throw_line(""); }
    
    // First: lets count the number of rows
    UInt number_of_rows = 0;
    UInt number_of_columns = 0;
    while (std::getline(in_file, line)) {
      std::vector<std::string> elements = Split(line, '\t');
      if (number_of_columns == 0) { number_of_columns = elements.size(); }
      else {
        if (number_of_columns != elements.size()) { throw_line(""); }
      }
      
      ++number_of_rows; 
    }
    
    // Reset pointer
    in_file.clear();
    in_file.seekg(0,std::ios::beg);
    
    // Create new matrix
    // TODO: recognize complex matrix
    Matrix<T> matrix(number_of_rows, number_of_columns);
    for(UInt row=0; row<number_of_rows; ++row) {
      std::getline(in_file, line);
      std::vector<std::string> elements = Split(line, '\t');
      for (UInt column=0; column<elements.size(); ++column) {
        matrix.set_element(row, column, (T) StringToDouble(elements[column]));
      }
    }
    
    in_file.close();
    return matrix;
  }
  
  /**
   Constructs a matrix of all ones. Equivalent to Matlab's ones(N,M).
   */
  static Matrix Ones(UInt number_of_rows, UInt number_of_columns) {
    Matrix<T> matrix(number_of_rows, number_of_columns);
    for(UInt row=0; row<number_of_rows; ++row) {
      for (UInt column=0; column<number_of_columns; ++column) {
        matrix.set_element(row, column, (T) 1.0);
      }
    }
    return matrix;
  }

  /**
   Constructs a matrix of all ones. Equivalent to Matlab's ones(N).
   */
  static Matrix Ones(UInt matrix_dimension) {
    return Matrix<T>::Ones(matrix_dimension, matrix_dimension);
  }
  
private:
  // Outer is rows, inner is columns. Hence, data_[0] is the first column.
  std::vector<std::vector<T> > data_;
  UInt num_columns_;
  UInt num_rows_;
};
  
template<class T>
MCL_API void Print(const Matrix<T>& matrix) {
  for (UInt i=0; i<matrix.num_rows(); ++i) {
    for (UInt j=0; j<matrix.num_columns(); ++j) {
      std::cout<<matrix.element(i,j)<<"\t";
    }
    std::cout<<std::endl;
  }
}

/** Transposes the matrix. Equivalent to Matlab's matrix' */
template<class T>
MCL_API Matrix<T> Transpose(const Matrix<T>& matrix) {
  Matrix<T> output(matrix.num_columns(), matrix.num_rows());
  
  for (UInt i=0; i<output.num_rows(); ++i) {
    for (UInt j=0; j<output.num_columns(); ++j) {
      output.set_element(i, j, matrix.element(j, i));
    }
  }
  return output;
}
  
/** 
 Multiplies all the elements of `matrix` by `value`. Equivalent
 to Matlabs' matrix.*value
 */
template<class T>
MCL_API Matrix<T> Multiply(const Matrix<T>& matrix, T value) {
  Matrix<T> output(matrix.num_rows(), matrix.num_columns());
  for (UInt i=0; i<output.num_rows(); ++i) {
    for (UInt j=0; j<output.num_columns(); ++j) {
      output.set_element(i, j, matrix.element(i, j)*value);
    }
  }
  return output;
}
 
  
/** Matrix multiplication. Equivalent to Matlabs' matrix_a*matrix_b */
template<class T>
MCL_API Matrix<T> Multiply(const Matrix<T>& matrix_a,
                           const Matrix<T>& matrix_b) {
  if (matrix_a.num_columns() != matrix_b.num_rows()) { throw_line(""); }
  
  Matrix<T> output(matrix_a.num_rows(), matrix_b.num_columns());
  for (UInt i=0; i<output.num_rows(); ++i) {
    for (UInt j=0; j<output.num_columns(); ++j) {
      T output_value = (T) 0.0;
      for (UInt k=0; k<matrix_a.num_columns(); ++k) {
        output_value += matrix_a.element(i, k) * matrix_b.element(k, j);
      }
      output.set_element(i, j, output_value);
    }
  }
  return output;
}
  
template<class T>
MCL_API std::vector<T> Multiply(const Matrix<T>& matrix_a,
                                const std::vector<T>& vector) {
  if (matrix_a.num_columns() != vector.size()) { throw_line(""); }
  Matrix<T> temp_input(vector.size(), 1);
  temp_input.set_column(0, vector);
  Matrix<T> temp_output = Multiply(matrix_a, temp_input);
  if (temp_output.num_columns() != 1) { throw_line(""); }
  if (temp_output.num_rows() != vector.size()) { throw_line(""); }
  
  return temp_output.column(0);
}
  
  
/** 
 Extract the maximum value of the matrix. Equivalent to Matlab's
 max(max(matrix)) 
 */
template<class T>
MCL_API T Max(const Matrix<T>& matrix) {
  return Max<T>(matrix.Serial());
}
  
  
/** Contains eigenvalues and eigenvectors */
struct EigOutput {
  std::vector<Complex> eigen_values; /**< Eigenvalues */
  std::vector<std::vector<Complex> > eigen_vectors; /**< Eigenvectors */
};
  
MCL_API EigOutput Eig(const Matrix<Real>& matrix);

MCL_API Matrix<Real> RealPart(const Matrix<Complex>& input);
  
#if MCL_LOAD_EIGEN
MCL_API Eigen::MatrixXd ConvertToEigen(const Matrix<Real>& input);
#endif
  
template<class T>
MCL_API bool IsEqual(const Matrix<T>& matrix_a, const Matrix<T>& matrix_b) {
  if (matrix_a.num_rows() != matrix_b.num_rows() |
      matrix_a.num_columns() != matrix_b.num_columns()) { return false; }
  
  for (UInt i=0; i<matrix_a.num_rows(); ++i) {
    for (UInt j=0; j<matrix_a.num_columns(); ++j) {
      if (!IsEqual(matrix_a.element(i, j), matrix_b.element(i, j))) {
        return  false;
      }
    }
  }
  return true;
}

bool MatrixOpTest();
  
  
} // namespace mcl

#endif
